"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.generateClient = void 0;
const fs_extra_1 = __importDefault(require("fs-extra"));
const klaw_1 = __importDefault(require("klaw"));
const store_1 = require("./store");
function generateClient() {
    return __awaiter(this, void 0, void 0, function* () {
        // TODO we must run tsconfig.js with     "declaration": true,
        // this will generate .d.ts
        if (!process.env.CLIENT_MODE) {
            return;
        }
        setTimeout(() => __awaiter(this, void 0, void 0, function* () {
            const config = store_1.getConfig();
            console.log("Client Generation Started.");
            const controllerPath = `${process.cwd()}/${config.controllersPath}`.replace("dist", "src");
            const paths = yield getFilesForDir(controllerPath);
            const validPaths = paths.filter((path) => path.includes(".ts"));
            let fnString = "";
            let interfaceList = [];
            let interfaceString = "";
            for (const path of validPaths) {
                const file = yield fs_extra_1.default.readFile(path, "utf8");
                const routeDeclarations = yield findAllAddRouteOccurences(path);
                // console.log(path);
                const routeDatas = routeDeclarations.map((index) => findReturnedValue(file.slice(index)));
                yield findFunctionsData(routeDatas);
                fnString += yield composeClientFunctions(routeDatas);
                // console.log(routeDatas);
                let interfaces = composeInterfacesList(routeDatas);
                interfaceList = [...interfaceList, ...interfaces];
            }
            interfaceList = [...new Set(interfaceList)];
            // console.log(interfaceList);
            // interfaceString = await extractNeededTypesFromProject({ interfaceList }); ------> TODO: review the whole logic to import types.
            yield composeClientClass({ fnString, interfaceString });
        }), 1000);
    });
}
exports.generateClient = generateClient;
function extractNeededTypesFromProject({ interfaceList, passes = 0, }) {
    return __awaiter(this, void 0, void 0, function* () {
        // console.log("extractNeededTypesFromProject", interfaceList);
        let interfaceString = "";
        let interfaceStringList = [];
        // Find type location
        const projectPaths = yield getFilesForDir(`${process.cwd()}`);
        const projectValidPaths = projectPaths.filter((path) => path.includes("d.ts"));
        for (const path of projectValidPaths) {
            const file = yield fs_extra_1.default.readFile(path, "utf8");
            for (const type of interfaceList) {
                const includeType = file.includes(`interface ${type}`);
                if (includeType) {
                    const typeString = yield extractSpecificTypeFromFile(path, type);
                    if (!interfaceStringList.includes(typeString)) {
                        interfaceStringList.push(typeString);
                    }
                }
            }
        }
        // TODO find nested Types.
        for (const intstr of interfaceStringList) {
            const nestedTypes = getNestedTypesFromTypeString(intstr);
            // console.log("getNestedTypesFromTypeString", nestedTypes);
            if (nestedTypes.length &&
                !interfaceList.includes(intstr) &&
                !nestedTypes.includes(intstr) &&
                passes <= 3) {
                const nestedInterfaces = yield extractNeededTypesFromProject({
                    interfaceList: [...new Set([...nestedTypes, ...interfaceList])],
                    passes: passes + 1,
                });
                interfaceStringList = [
                    ...new Set([...interfaceStringList, nestedInterfaces]),
                ];
            }
        }
        interfaceStringList.forEach((intstr) => {
            interfaceString += intstr + "\n";
        });
        // console.log(interfaceStringList);
        return interfaceString;
    });
}
function getNestedTypesFromTypeString(type) {
    const nestedTypeName = [];
    let level = 0;
    let startIndex = type.indexOf("{");
    let isScaningType = false;
    let tempNestedType = "";
    for (let i = startIndex; i < type.length; i++) {
        if (type[i] === "{") {
            level++;
            isScaningType = false; // it's an inline type
        }
        if (type[i] === "}") {
            if (level === 0) {
                break; // end of type
            }
            else {
                level--;
            }
        }
        if (type[i] === ":") {
            isScaningType = true;
            // we found type declaration
        }
        else if (isScaningType) {
            if (type[i] === ";") {
                // end of type declaration;
                isScaningType = false;
                tempNestedType = tempNestedType.trim();
                tempNestedType = tempNestedType.replace("]", "");
                tempNestedType = tempNestedType.replace("[", "");
                if (!tempNestedType.includes("boolean") &&
                    !tempNestedType.includes("string") &&
                    !tempNestedType.includes("number") &&
                    !tempNestedType.includes("any") &&
                    !tempNestedType.includes("void")) {
                    nestedTypeName.push(tempNestedType);
                }
                tempNestedType = "";
            }
            else {
                tempNestedType += type[i];
            }
        }
    }
    return [...new Set([...nestedTypeName])];
}
function extractSpecificTypeFromFile(filePath, typeName) {
    return __awaiter(this, void 0, void 0, function* () {
        let finalStr = "";
        const fileString = yield fs_extra_1.default.readFile(filePath, "utf8");
        const typeStartIndex = fileString.indexOf("interface " + typeName);
        if (typeStartIndex === -1) {
            return finalStr;
        }
        let level = 0;
        let gotCore = false;
        for (let i = typeStartIndex; i < fileString.length; i++) {
            finalStr += fileString[i];
            if (fileString[i] === "{") {
                if (gotCore) {
                    level += 1;
                }
                else {
                    gotCore = true;
                }
            }
            if (fileString[i] === "}") {
                if (level === 0) {
                    break;
                }
                else {
                    level -= 1;
                }
            }
        }
        if (finalStr.length) {
            return finalStr;
        }
        return finalStr;
    });
}
function composeInterfacesList(routeDatas) {
    let interfaces = [];
    routeDatas.forEach((route) => {
        if (route.returnedType.includes("Promise<")) {
            const returnedType = route.returnedType;
            const startIndex = returnedType.indexOf("Promise<") + 8;
            let interfaceName = "";
            for (let i = startIndex; i < returnedType.length; i++) {
                if (returnedType[i] === "{" || returnedType[i] === "[") {
                    break; // it's an inline type
                }
                if (returnedType[i] === ">") {
                    break; // it's the end
                }
                interfaceName += returnedType[i];
            }
            interfaceName = interfaceName.replace("[]", "");
            interfaceName = interfaceName.trim();
            if (interfaceName.length &&
                interfaceName !== "void" &&
                interfaceName !== "any") {
                interfaces.push(interfaceName);
            }
        }
    });
    interfaces = [...new Set(interfaces)];
    return interfaces;
}
function composeClientClass({ fnString, interfaceString, }) {
    return __awaiter(this, void 0, void 0, function* () {
        let destination = "./sexpress/";
        if (process.env.CLIENT_DESTINATION) {
            destination = process.env.CLIENT_DESTINATION;
        }
        yield fs_extra_1.default.ensureDir(destination);
        let classTemplate = yield fs_extra_1.default.readFile(`${process.cwd()}/node_modules/@appandflow/s-express/dist/lib/clientClassTemplate.txt`, "utf8");
        let exportTemplate = yield fs_extra_1.default.readFile(`${process.cwd()}/node_modules/@appandflow/s-express/dist/lib/clientExportTemplate.txt`, "utf8");
        classTemplate = classTemplate.replace("FUNCTIONS", fnString);
        classTemplate = classTemplate.replace("// TYPES_CLIENT", interfaceString);
        if (process.env.CLIENT_TYPE_PATH) {
            classTemplate = classTemplate.replace("// TYPES_IMPORT", process.env.CLIENT_TYPE_PATH);
        }
        else {
            classTemplate = classTemplate.replace("// TYPES_IMPORT", "");
        }
        yield fs_extra_1.default.writeFile(`${destination}/sexpressClass.ts`, classTemplate);
        const exists = yield fs_extra_1.default.pathExists(`${destination}/index.ts`);
        if (!exists) {
            yield fs_extra_1.default.writeFile(`${destination}/index.ts`, exportTemplate);
        }
        console.log(`Client generated at ${destination}`);
        process.exit();
    });
}
function composeClientFunctions(routeDatas) {
    return __awaiter(this, void 0, void 0, function* () {
        const fnTemplate = yield fs_extra_1.default.readFile(`${process.cwd()}/node_modules/@appandflow/s-express/dist/lib/clientFunctionTemplate.txt`, "utf8");
        let importPrefx = "";
        if (process.env.CLIENT_TYPE_PATH) {
            importPrefx = "ALL_TYPES.";
        }
        let fnsString = "\n";
        routeDatas.forEach((routeData) => {
            let fnText = fnTemplate;
            routeData.returnedType = correctTypeIfClientTypePathIsDefined(routeData.returnedType);
            fnText = fnText.replace("RETURNED_TYPE", routeData.returnedType);
            fnText = fnText.replace("HTTP_METHOD", routeData.httpMethod.toLowerCase());
            fnText = fnText.replace("HTTP_METHOD", routeData.httpMethod.toLowerCase());
            fnText = fnText.replace("ROUTE_URL", routeData.urlPath);
            if (routeData.payloadType === "void" || !routeData.payloadType) {
                fnText = fnText.replace("payload: PAYLOAD_TYPE", "");
                fnText = fnText.replace(", PAYLOAD_TO_SEND", "");
            }
            else {
                routeData.payloadType = correctTypeIfClientTypePathIsDefined(routeData.payloadType);
                fnText = fnText.replace("PAYLOAD_TYPE", routeData.payloadType);
                fnText = fnText.replace("PAYLOAD_TO_SEND", "payload");
            }
            let tempName = "";
            let fnName = `${routeData.httpMethod.toLowerCase()}`;
            let nextIsCapital = false;
            for (let i = 0; i < routeData.urlPath.length; i++) {
                if (routeData.urlPath[i] === "/") {
                    nextIsCapital = true;
                }
                else {
                    if (nextIsCapital) {
                        tempName += routeData.urlPath[i].toUpperCase();
                        nextIsCapital = false;
                    }
                    else {
                        tempName += routeData.urlPath[i];
                    }
                }
            }
            // fnText = fnText.replace("FUNCTION_NAME", `${fnName}${tempName}`); // lets use the function name for now
            fnText = fnText.replace("FUNCTION_NAME", routeData.string);
            fnsString += fnText + "\n";
        });
        return fnsString;
    });
}
function findFunctionsData(routeDatas) {
    return __awaiter(this, void 0, void 0, function* () {
        // We now need to find the function .d.ts in the dist folder
        const distPath = `${process.cwd()}/dist`; // TODO add support for custom dist
        const paths = yield getFilesForDir(distPath);
        const validPaths = paths.filter((path) => path.includes(".d.ts"));
        for (const path of validPaths) {
            for (const routeData of routeDatas) {
                yield extractReturnedValueFromFunction(path, routeData);
            }
        }
    });
}
function extractReturnedValueFromFunction(filePath, routeData) {
    return __awaiter(this, void 0, void 0, function* () {
        const file = yield fs_extra_1.default.readFile(filePath, "utf8");
        // We add '(' so we know it the end of the function declaration
        const fnIndex = file.indexOf(routeData.string + "(");
        if (fnIndex !== -1) {
            const fnStart = file.substring(fnIndex);
            let startTypeIndex = fnStart.indexOf("):");
            if (startTypeIndex !== -1) {
                startTypeIndex += 2;
                const isPromise = fnStart
                    .substring(startTypeIndex, startTypeIndex + 15)
                    .includes("Promise<");
                if (isPromise) {
                    const type = fnStart.substring(startTypeIndex, fnStart.indexOf(">;")) + ">";
                    routeData.returnedType = type;
                }
            }
        }
    });
}
function findAllAddRouteOccurences(filePath) {
    return __awaiter(this, void 0, void 0, function* () {
        const file = yield fs_extra_1.default.readFile(filePath, "utf8");
        const addRouteLocations = [
            ...searchIndexes(file, "addRoute<"),
            ...searchIndexes(file, "addRoute("),
        ]; // Todo add support for other thing then "addRoute"
        return addRouteLocations;
    });
}
function getFilesForDir(dirPath) {
    return __awaiter(this, void 0, void 0, function* () {
        return new Promise((resolve) => {
            const items = [];
            klaw_1.default(dirPath, { depthLimit: -1 })
                .on("data", (item) => items.push(item.path))
                .on("end", () => resolve(items));
        });
    });
}
function findReturnedValue(str) {
    const res = {
        string: "",
        type: "object",
        isImported: false,
        importLocation: "",
        isComposed: false,
        composedString: "",
        urlPath: "/",
        httpMethod: "GET",
        returnedType: "void",
        payloadType: "void",
    };
    const returnIndex = str.indexOf("return");
    let valueStartIndex = 0;
    if (returnIndex !== -1) {
        valueStartIndex = returnIndex + 7; // "return " string index
        if (str[valueStartIndex + 1] === "{") {
            res.type = "object";
            // todo handle object
        }
        else {
            // function/variable case
            // i.e: "service.function" or "function" or "variable"
            let string = "";
            for (let i = valueStartIndex; i < str.length; i++) {
                if (str[i] === "(") {
                    // it's a function
                    res.type = "function";
                    res.string = string;
                    res.isImported = true; // TODO for now we will suppose it's imported
                    break;
                }
                else if (str[i] === ".") {
                    res.isComposed = true;
                    res.composedString = string; // todo handle multiple level of composition i.e: services.a.b.c
                    string = "";
                }
                else {
                    string += str[i];
                }
            }
        }
        // now we need to figure out the url path
        const pathStartIndex = str.indexOf("path");
        if (pathStartIndex !== -1) {
            const startIndex = pathStartIndex + 5; // +path.length + ":"
            let stringStartPosition = -1;
            let endChar;
            for (let i = startIndex; i < str.length; i++) {
                if ((str[i] === '"' || str[i] === "`" || str[i] === "'") && !endChar) {
                    // start
                    endChar = str[i];
                    stringStartPosition = i + 1;
                }
                else if (endChar) {
                    if (str[i] === endChar) {
                        res.urlPath = str.substring(stringStartPosition, i);
                        break;
                    }
                }
            }
        }
        // now we need to figure out the http method
        const methodStartIndex = str.indexOf("method");
        if (methodStartIndex !== -1) {
            const startIndex = methodStartIndex + 7; // +method.length + ":"
            let stringStartPosition = -1;
            let endChar;
            for (let i = startIndex; i < str.length; i++) {
                if ((str[i] === '"' || str[i] === "`" || str[i] === "'") && !endChar) {
                    // start
                    endChar = str[i];
                    stringStartPosition = i + 1;
                }
                else if (endChar) {
                    if (str[i] === endChar) {
                        res.httpMethod = str.substring(stringStartPosition, i);
                        break;
                    }
                }
            }
        }
        // Now we will figure out the payload type
        const startIndex = str.indexOf("addRoute<"); // we only support addRoute for now.
        if (startIndex !== -1) {
            let payloadType = "";
            let isInline = false;
            let level = 0;
            for (let i = startIndex + "addRoute<".length; i < str.length; i++) {
                if (str[i] === "(") {
                    // instant break it's void type.
                    break;
                }
                payloadType += str[i];
                if (str[i] === "{") {
                    if (!isInline) {
                        isInline = true;
                    }
                    else {
                        level += 1;
                    }
                }
                if (str[i] === "}" && isInline) {
                    if (level !== 0) {
                        level -= 1;
                    }
                    else {
                        break; // end of inline object
                    }
                }
                if (!isInline && str[i] !== "{") {
                    if (str[i] === ">" || str[i] === ",") {
                        if (str[i] === ",") {
                            payloadType = payloadType.substring(0, payloadType.length - 1);
                        }
                        // we wont support nested generic type for now i.e: A<B<c>> // TODO add support for partial.
                        break;
                    }
                }
            }
            if (payloadType.length) {
                res.payloadType = payloadType;
            }
        }
    }
    return res;
}
function searchIndexes(source, find) {
    if (!source) {
        return [];
    }
    // if find is empty string return all indexes.
    if (!find) {
        // or shorter arrow function:
        // return source.split('').map((_,i) => i);
        return source.split("").map(function (_, i) {
            return i;
        });
    }
    const result = [];
    for (let i = 0; i < source.length; ++i) {
        // If you want to search case insensitive use
        // if (source.substring(i, i + find.length).toLowerCase() == find) {
        if (source.substring(i, i + find.length) == find) {
            result.push(i);
        }
    }
    return result;
}
function correctTypeIfClientTypePathIsDefined(typeString) {
    let mustAddPromise = false;
    if (!process.env.CLIENT_TYPE_PATH) {
        return typeString;
    }
    if (typeString.includes("Promise<")) {
        mustAddPromise = true;
        typeString = typeString.replace("Promise<", "");
        typeString = typeString.replace(">", "");
    }
    const primatyTypes = [
        "number",
        "string",
        "null",
        "void",
        "undefined",
        "any",
        "unknown",
        "boolean",
    ];
    const correctedTypes = [];
    const extractedTypes = extractTypesFromObj(typeString);
    extractedTypes.forEach((types) => {
        let corrected = types;
        if (primatyTypes.includes(types)) {
            correctedTypes.push(corrected);
        }
        else {
            let edited = "";
            const words = types.split(" ");
            const correctedWords = words.map((word) => {
                // fix for (
                let needFirstParenthese = false;
                if (word.charAt(0) === "(") {
                    needFirstParenthese = true;
                    word = word.slice(1);
                }
                if (primatyTypes.includes(word)) {
                    return word;
                }
                else if (word.length === 1 && !isLetter(word)) {
                    return word;
                }
                else {
                    if (word.includes("Partial<") && word.includes(">")) {
                        let temp = "";
                        temp = word.replace("Partial<", "");
                        temp = temp.replace(">", "");
                        return `Partial<ALL_TYPES.${temp}>`;
                    }
                    else {
                        if (word.charAt(0) === "'" || word.charAt(0) === '"') {
                            // enum string.
                            return word;
                        }
                        else {
                            return `${needFirstParenthese ? "(" : ""}ALL_TYPES.${word}`;
                        }
                    }
                }
            });
            edited = correctedWords.join(" ");
            correctedTypes.push(edited.trim());
        }
    });
    let correctedVersion = typeString;
    extractedTypes.forEach((extracted, index) => {
        if (!correctedVersion.includes(`ALL_TYPES.${extracted}`)) {
            correctedVersion = correctedVersion.replaceAll(extracted, correctedTypes[index]);
        }
    });
    if (mustAddPromise) {
        correctedVersion = `Promise<${correctedVersion}>`;
    }
    return correctedVersion;
}
function isLetter(str) {
    return str.length === 1 && str.match(/[a-z]/i);
}
/**
 *
 * Receive { videos: Artist['videos']; text: string }
 * and returns
 * ["Artist['videos']", "string"]
 */
function extractTypesFromObj(obj) {
    let types = [];
    let started = false;
    let currentType = "";
    let typeDeclarationStarted = false;
    for (let i = 0; i < obj.length; i++) {
        if (obj.charAt(i) === "}") {
            started = false;
            currentType = currentType.trim();
            if (currentType.length) {
                types.push(currentType);
            }
            break;
        }
        else if (obj.charAt(i) === "{") {
            started = true;
        }
        else if (started) {
            if (obj.charAt(i) === ":") {
                typeDeclarationStarted = true;
            }
            else if (typeDeclarationStarted) {
                currentType += obj.charAt(i);
                if (obj.charAt(i) === ";" || obj.charAt(i) === ",") {
                    typeDeclarationStarted = false;
                    types.push(currentType.trim());
                    currentType = "";
                }
            }
        }
    }
    if (!types.length && obj.length) {
        types.push(obj.trim());
    }
    types = types.map((type) => {
        return type.replace(";", "");
    });
    return types;
}
